﻿using System;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Windows.Media.Imaging;
using System.Xml;
using System.Text;

namespace WpfGeoTagger.Models
{
	/// <summary>
	/// GeoTag を管理する為のクラスです。
	/// </summary>
	class GeoTag
	{
		/// <summary>
		/// 現在の緯度・経度を使用して、標高を更新します。
		/// 更新に失敗した場合、標高は 0.0 に設定されます。
		/// <para>
		/// 標高は USGS ( アメリカ地質調査所 ) の getElevation サービスから取得しています。
		/// USGS は全世界から参照される可能性があるので、サービスの負荷を軽減する為に、呼び出し頻度が少なくなるようにしてください。
		/// </para>
		/// </summary>
		public void UpdateAltitude()
		{
			this.Altitude = 0.0;

			try
			{
				// getElevation は X_Value に経度、Y_Value に緯度を指定する
				var url = String.Format( "http://gisdata.usgs.gov/xmlwebservices2/elevation_service.asmx/getElevation?X_Value={0}&Y_Value={1}&Elevation_Units=METERS&Source_Layer=-1&Elevation_Only=true", this.Longitude, this.Latitude );
				var xml = new XmlDocument();

				xml.Load( url );

				var element = xml.SelectSingleNode( "//double" );
				if( element != null )
				{
					// 不正な緯度・経度を指定した場合は "-1.79769313486231E+308" という文字列が入り、
					// Convert.ToDouble に失敗するので、必ず例外をハンドリングしておく事。
					//
					this.Altitude = Convert.ToDouble( element.InnerText );
				}
			}
			catch( Exception exp )
			{
				Debug.WriteLine( exp.Message );
			}
		}

		/// <summary>
		/// 指定した Object が、現在の Object と等しいかどうかを判断します。
		/// </summary>
		/// <param name="obj">現在の Object と比較する Object。</param>
		/// <returns>指定した Object が現在の Object と等しい場合は true。それ以外の場合は false。</returns>
		public override bool Equals( object obj )
		{
			if( obj == null || this.GetType() != obj.GetType() ) { return false; }

			var geoTag = obj as GeoTag;
			if( this.Latitude  != geoTag.Latitude  ||
				this.Longitude != geoTag.Longitude || 
				this.Altitude  != geoTag.Altitude )
			{
				return false;
			}

			return true;
		}

		/// <summary>
		/// 特定の型のハッシュ関数として機能します。
		/// </summary>
		/// <returns>現在の Object のハッシュ コード。</returns>
		public override int GetHashCode()
		{
			return this.Latitude.GetHashCode() ^ this.Longitude.GetHashCode() ^ this.Altitude.GetHashCode();
		}

		/// <summary>
		/// 現在の GeoTag を EXIF の GPS としてメタデータへ書き込みます。
		/// </summary>
		/// <param name="metadata">画像のメタデータを表すオブジェクト。</param>
		public void Write( BitmapMetadata metadata )
		{
			// GPS のバージョン情報
			{
				var version = new byte[ 4 ] { 2, 0, 0, 0 };
				metadata.SetQuery( GeoTag.ExifPathGpsVersion, version );
			}

			// 緯度
			{
				var values = GeoTag.ConvertToGpsMultiValues( this.Latitude );
				metadata.SetQuery( GeoTag.ExifPathGpsLatitude, values );

				var direction = ( this.Latitude >= 0 ? "N" : "S" );
				metadata.SetQuery( GeoTag.ExifPathGpsLatitudeDirection, direction );
			}

			// 経度
			{
				var values = GeoTag.ConvertToGpsMultiValues( this.Longitude );
				metadata.SetQuery( GeoTag.ExifPathGpsLongitude, values );

				var direction = ( this.Longitude >= 0 ? "E" : "W" );
				metadata.SetQuery( GeoTag.ExifPathGpsLongitudeDirection, direction );
			}

			// 標高
			{
				var value = GeoTag.ConvertToGpsValue( this.Altitude );
				metadata.SetQuery( GeoTag.ExifPathGpsAltitude, value );
				metadata.SetQuery( GeoTag.ExifPathGpsAltitudeBase, ( byte )0 );
			}

			// 測地系 ( Google Map の場合は「世界測地系 ( WGS84 )」となる )
			{
				metadata.SetQuery( GeoTag.ExifPathGpsMapDatum, "WGS84" );
			}
		}

		/// <summary>
		/// 現在の GeoTag を表す System.String を返します。
		/// </summary>
		/// <returns>GeoTag の内容を表す System.String。</returns>
		public override string ToString()
		{
			if( this.Altitude == 0.0 )
			{
				return String.Format( "緯度 = {0}、経度 = {1}", this.Latitude, this.Longitude );
			}
			else
			{
				return String.Format( "緯度 = {0}、経度 = {1}、標高 = {2}m", this.Latitude, this.Longitude, this.Altitude );
			}
		}

		#region 静的メソッド

		/// <summary>
		/// 画像のメタデータから GeoTag インスタンスを生成します。
		/// </summary>
		/// <param name="metadata">画像のメタデータ。</param>
		/// <returns>成功時は GeoTag インスタンス。失敗時は null 参照。</returns>
		public static GeoTag FromMetadata( BitmapMetadata metadata )
		{
			// 標高
			double altitude = 0.0;
			{
				var value = metadata.GetQuery( GeoTag.ExifPathGpsAltitude ) as ulong?;
				if( value != null )
				{
					altitude = GeoTag.ConvertToGeoTagValue( value.Value );
				}
			}

			// 緯度
			double latitude = 0.0;
			{
				var gpsLatitude = metadata.GetQuery( GeoTag.ExifPathGpsLatitude ) as ulong[];
				if( gpsLatitude == null ) { return null; }

				var direction = metadata.GetQuery( GeoTag.ExifPathGpsLatitudeDirection ) as string;
				if( direction == null ) { return null; }

				// 南半球なら負の値
				latitude = GeoTag.ConvertToGeoTagLocation( gpsLatitude );
				if( direction == "S" )
				{
					latitude *= -1;
				}
			}

			// 経度
			double longitude = 0.0;
			{
				var gpsLongitude = metadata.GetQuery( GeoTag.ExifPathGpsLongitude ) as ulong[];
				if( gpsLongitude == null ) { return null; }

				var direction = metadata.GetQuery( GeoTag.ExifPathGpsLongitudeDirection ) as string;
				if( direction == null ) { return null; }

				// 西半球なら負の値
				longitude = GeoTag.ConvertToGeoTagLocation( gpsLongitude );
				if( direction == "W" )
				{
					longitude *= -1;
				}
			}

			return new GeoTag() { Latitude = latitude, Longitude = longitude, Altitude = altitude };
		}

		/// <summary>
		/// EXIF から得られた GPS 情報を表す { 度, 分, 秒 } の配列から GeoTag 用の位置情報を取得します。 
		/// </summary>
		/// <param name="gpsValues">GPS 情報の配列。</param>
		/// <returns>GeoTag 用の位置情報。</returns>
		private static double ConvertToGeoTagLocation( ulong[] gpsValues )
		{
			double degrees = GeoTag.ConvertToGeoTagValue( gpsValues[ 0 ] );
			double minutes = GeoTag.ConvertToGeoTagValue( gpsValues[ 1 ] );
			double seconds = GeoTag.ConvertToGeoTagValue( gpsValues[ 2 ] );

			// GeoTag の位置へ変換
			return degrees + ( minutes / 60.0 ) + ( seconds / 3600 );
		}

		/// <summary>
		/// GeoTag の位置情報から EXIF 用の GPS 情報を取得します。
		/// </summary>
		/// <param name="location">GeoTag の位置情報。</param>
		/// <returns>EXIF 用の GPS 情報。</returns>
		private static ulong[] ConvertToGpsMultiValues( double location )
		{
			location = Math.Abs( location );

			// 度
			double degrees = Math.Floor( location );
			location -= degrees;

			// 分
			double minutes = Math.Floor( location * 60.0 );
			location -= ( minutes / 60.0 );

			// 秒
			double seconds = Math.Round( location * 3600.0 * 1000000.0 );

			// GPS 情報へ変換する
			ulong[] values = new ulong[ 3 ];
			values[ 0 ] = GeoTag.ConvertToGpsValue( degrees );
			values[ 1 ] = GeoTag.ConvertToGpsValue( minutes );
			values[ 2 ] = GeoTag.ConvertToGpsValueSeconds( seconds );

			return values;
		}

		/// <summary>
		/// GeoTag の値を GPS 情報の値へ変換します。
		/// ただし緯度・経度の秒は特殊なので、変換には ConvertToGpsValueSeconds を利用して下さい。
		/// </summary>
		/// <param name="value">GeoTag の値。</param>
		/// <returns>変換された値。</returns>
		private static ulong ConvertToGpsValue( double value )
		{
			return Convert.ToUInt64( value + 0x100000000 );
		}

		/// <summary>
		/// GeoTag の秒データから GPS 用の秒データを取得します。
		/// </summary>
		/// <param name="seconds">GeoTag の秒。</param>
		/// <returns>作成した GPS 用の秒データ。</returns>
		private static ulong ConvertToGpsValueSeconds( double seconds )
		{
			byte[] num = BitConverter.GetBytes( ( int )seconds );
			byte[] den = BitConverter.GetBytes( ( int )1000000 );
			byte[] sec = new byte[ 8 ] { num[ 0 ], num[ 1 ], num[ 2 ], num[ 3 ], den[ 0 ], den[ 1 ], den[ 2 ], den[ 3 ] };

			return BitConverter.ToUInt64( sec, 0 );
		}

		/// <summary>
		/// GPS 情報の一つのデータ ( GPSLatitude、GPSLongitude、GPSAltitude ) を GeoTag の値へ変換します。
		/// EXIF のタグ情報については、以下のページを参照して下さい。
		/// <span>
		/// http://homepage1.nifty.com/gigo/DC/GPS/gpsifd.html
		/// </span>
		/// </summary>
		/// <param name="value">GPS 情報の値。</param>
		/// <returns>変換された値。</returns>
		private static double ConvertToGeoTagValue( ulong value )
		{
			byte[]	bytes = BitConverter.GetBytes( value );
			int		upper = BitConverter.ToInt32( bytes, 0 );
			int		lower = BitConverter.ToInt32( bytes, 4 );

			return ( ( double )upper / ( double )lower );
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// 標高を取得または設定します。
		/// </summary>
		public double Altitude { get; set; }

		/// <summary>
		/// 緯度を取得または設定します。
		/// </summary>
		public double Latitude { get; set; }

		/// <summary>
		/// 経度を取得または設定します。
		/// </summary>
		public double Longitude { get; set; }

		#endregion

		#region 定数

		/// <summary>
		/// EXIF の GPS 情報における「バージョン」を示すパス文字列。
		/// </summary>
		private const string ExifPathGpsVersion = "/app1/ifd/gps/";

		/// <summary>
		/// EXIF の GPS 情報における「測地系」を示すパス文字列。
		/// </summary>
		private const string ExifPathGpsMapDatum = "/app1/ifd/gps/subifd:{short=18}";

		/// <summary>
		/// EXIF の GPS 情報における「標高の基準点のモード ( 0 = 海抜、1 = それ以外 )」を示すパス文字列。
		/// </summary>
		private const string ExifPathGpsAltitudeBase = "/app1/ifd/gps/subifd:{ushort=5}";

		/// <summary>
		/// EXIF の GPS 情報における「標高」を示すパス文字列。
		/// </summary>
		private const string ExifPathGpsAltitude = "/app1/ifd/gps/subifd:{ulong=6}";

		/// <summary>
		/// EXIF の GPS 情報における「緯度」を示すパス文字列。
		/// </summary>
		private const string ExifPathGpsLatitude = "/app1/ifd/gps/subifd:{ulong=2}";

		/// <summary>
		/// EXIF の GPS 情報における「緯度の進行方向 ( "N" または "S" )」を示すパス文字列。
		/// </summary>
		private const string ExifPathGpsLatitudeDirection = "/app1/ifd/gps/subifd:{short=1}";

		/// <summary>
		/// EXIF の GPS 情報における「緯度」を示すパス文字列。
		/// </summary>
		private const string ExifPathGpsLongitude = "/app1/ifd/gps/subifd:{ulong=4}";

		/// <summary>
		/// EXIF の GPS 情報における「経度の進行方向 ( "E" または "W" )」を示すパス文字列。
		/// </summary>
		private const string ExifPathGpsLongitudeDirection = "/app1/ifd/gps/subifd:{short=3}";

		#endregion
	}
}
