﻿using System;
using System.ComponentModel;
using System.IO;
using System.Windows.Media.Imaging;
using WpfGeoTagger.Models;

namespace WpfGeoTagger.ViewModels
{
	/// <summary>
	/// Image の Model と View を仲介します。
	/// </summary>
	class ImageViewModel : ViewModelBase
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		/// <param name="id">画像の識別子。</param>
		/// <param name="fileName">画像ファイルのパス。</param>
		/// <param name="onGeoTagLoadedd">GeoTag の読み込みが完了した事を通知するイベント ハンドラ。</param>
		/// <exception cref="FileNotFoundException">画像ファイルが存在しません。</exception>
		public ImageViewModel( int id, string fileName, Action< ImageViewModel > onGeoTagLoadedd )
		{
			if( !File.Exists( fileName ) ) { throw new FileNotFoundException( fileName ); }

			this.Id        = id;
			this.Name      = Path.GetFileName( fileName );
			this._fileName = fileName;
			
			this.LoadImage( fileName );
			this.LoadMetadata( fileName, onGeoTagLoadedd );
		}

		/// <summary>
		/// 画像からメタデータを読み込みます。
		/// </summary>
		/// <param name="fileName">画像ファイルのパス。</param>
		/// <param name="onMetadataLoadedd">メタデータの読み込みが完了した事を通知するイベント ハンドラ。</param>
		private void LoadMetadata( string fileName, Action< ImageViewModel > onMetadataLoadedd )
		{
			if( !File.Exists( fileName ) ) { return; }

			BackgroundWorker worker = new BackgroundWorker();
	
			worker.DoWork += ( sender, e ) =>
			{
				e.Result = MetadataViewModel.FromImage( fileName );
			};
		
			worker.RunWorkerCompleted += ( sender, e ) =>
			{
				var m = e.Result as MetadataViewModel;
				if( m != null )
				{
					this.Metadata = m;
					this.NotifyPropertyChanged( "Metadata" );

					if( onMetadataLoadedd != null )
					{
						onMetadataLoadedd( this );
					}
				}
			};

			worker.RunWorkerAsync();
		}

		/// <summary>
		/// 画像を読み込みます。
		/// </summary>
		/// <param name="fileName">画像ファイルのパス。</param>
		private void LoadImage( string fileName )
		{
			if( File.Exists( fileName ) )
			{
				BackgroundWorker worker = new BackgroundWorker();

				worker.DoWork += ( sender, e ) =>
				{
					BitmapImage image = new BitmapImage();
					image.BeginInit();
					image.DecodePixelWidth = 64;
					image.CacheOption = BitmapCacheOption.OnLoad;
					image.UriSource = new Uri( fileName );
					image.EndInit();
					image.Freeze();

					e.Result = image;
				};

				worker.RunWorkerCompleted += ( sender, e ) =>
				{
					var image = e.Result as BitmapImage;
					if( image != null )
					{
						this.Image = image;
						this.NotifyPropertyChanged( "Image" );
					}
				};

				worker.RunWorkerAsync();
			}
		}

		/// <summary>
		/// 現在のメタデータを画像ファイルに保存します。
		/// </summary>
		public void MetadataSave()
		{
			if( this.Metadata != null )
			{
				this.Metadata.Save( this._fileName );
			}
		}

		/// <summary>
		/// メタデータを設定または取得します。
		/// </summary>
		public MetadataViewModel Metadata { get; set; }
		
		/// <summary>
		/// 識別子を取得します。
		/// </summary>
		public int Id { get; private set; }

		/// <summary>
		/// 画像を設定または取得します。
		/// </summary>
		public BitmapImage Image { get; private set; }

		/// <summary>
		/// 画像の編集が行われた事を示す値を取得します。
		/// </summary>
		public bool IsEdited { get { return ( this.Metadata == null ? false : this.Metadata.IsEdited ); } }

		/// <summary>
		/// 画像ファイル名を取得します。
		/// </summary>
		public string Name { get; private set; }

		/// <summary>
		/// 画像ファイルのパス文字列。
		/// </summary>
		private string _fileName;
	}
}
