﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Media.Imaging;
using System.Xml;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using WpfGeoTagger.Models;

namespace WpfGeoTagger.ViewModels
{
	/// <summary>
	/// 画像の付加情報を格納します。
	/// </summary>
	class MetadataViewModel
	{
		/// <summary>
		/// 画像ファイルから Metadata インスタンスを生成します。
		/// </summary>
		/// <param name="fileName">画像ファイルのパス。</param>
		/// <returns>成功時は Metadata インスタンス。失敗時は null 参照。</returns>
		public static MetadataViewModel FromImage( string fileName )
		{
			using( var stream = new FileStream( fileName, FileMode.Open, FileAccess.Read, FileShare.Read ) )
			{
				var decoder   = new JpegBitmapDecoder( stream, BitmapCreateOptions.None, BitmapCacheOption.None );
				var metadata  = ( BitmapMetadata )decoder.Frames[ 0 ].Metadata;

				return new MetadataViewModel() { GeoTag = GeoTag.FromMetadata( metadata ), Address = Address.FromMetadata( metadata ) };
			}
		}

		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		private MetadataViewModel()
		{
		}

		/// <summary>
		///  メタデータの編集を開始する。
		/// </summary>
		public void BeginEdit()
		{
			this._address.BeginEdit();
			this._geoTag.BeginEdit();

			this.GeoTag = new GeoTag() { Altitude = this.GeoTag.Altitude, Latitude = this.GeoTag.Latitude, Longitude = this.GeoTag.Longitude };
			this.Address = null;

			// 編集中に標高が表示されないように、無効化しておく
			if( this.GeoTag != null )
			{
				this.GeoTag.Altitude = 0.0;
			}
		}

		/// <summary>
		/// メタデータの GeoTag 編集を終了する。
		/// </summary>
		public void FinishGeoTagEdit()
		{
			// 更新チェック
			var backup = this._geoTag.Backup;
			if( !this.GeoTag.Equals( backup ) )
			{
				this._isEditedGeoTag = true;
			}

			this._geoTag.FinishEdit( true );

			// 標高更新
			if( this.GeoTag != null )
			{
				this.GeoTag.UpdateAltitude();
			}
		}

		/// <summary>
		/// メタデータの住所編集を終了する。
		/// </summary>
		public void FinishAddressEdit( string address )
		{
			// 更新チェック
			var backup = this._address.Backup;
			var newAddress = Address.FromGeocoderAddress( address );
			if( ( backup == null && newAddress != null ) || ( backup != null && newAddress == null ) || !backup.Equals( newAddress ) )
			{
				this._isEditedAddress = true;
			}

			this.Address = newAddress;
			this._address.FinishEdit( true );
		}

		/// <summary>
		/// GeoTag を削除します。
		/// </summary>
		public void RemoveGeoTag()
		{
			this.GeoTag = null;
		}

		/// <summary>
		/// メタデータを画像へ保存します。
		/// </summary>
		/// <param name="fileName">画像ファイルのパス。</param>
		public void Save( string fileName )
		{
			if( !this.IsEdited ) { return; }

			using( var streamIn = new MemoryStream( File.ReadAllBytes( fileName ) ) )
			{
				var decoder = new JpegBitmapDecoder( streamIn, BitmapCreateOptions.PreservePixelFormat, BitmapCacheOption.None );

				using( var streamOut = File.Open( fileName, System.IO.FileMode.Open, System.IO.FileAccess.ReadWrite ) )
				{
					var encoder = new JpegBitmapEncoder();
					var frame   = BitmapFrame.Create( decoder.Frames[ 0 ] );

					encoder.Frames.Add( frame );

					var metadata = frame.Metadata as BitmapMetadata;

					// 編集アプリ名など ( アセンブリのタイトルをアプリ名とする )
					{
						var attribute = Attribute.GetCustomAttribute( Assembly.GetExecutingAssembly(), typeof( AssemblyTitleAttribute ) ) as AssemblyTitleAttribute;
						metadata.ApplicationName = attribute.Title;

						if( String.IsNullOrEmpty( metadata.CameraManufacturer ) )
						{
							metadata.CameraManufacturer = attribute.Title;
						}

						if( String.IsNullOrEmpty( metadata.CameraModel ) )
						{
							metadata.CameraModel = "PC";
						}
					}

					// EXIF バージョン ( 型が無いので BitmapMetadataBlob によるバイナリ書き込みを行う )
					{
						BitmapMetadataBlob version = new BitmapMetadataBlob( new byte[ 4 ] { 48, 50, 50, 49 } );
						metadata.SetQuery( MetadataViewModel.ExifPathVersion, version );
					}
					
					// 日時
					this.WriteDateTime( metadata );

					// 住所
					if( this._isEditedAddress && this.Address != null )
					{
						this.Address.Write( metadata );
					}

					// GeoTag
					if( this._isEditedGeoTag && this.GeoTag != null )
					{
						this.GeoTag.Write( metadata );
					}

					encoder.Frames.Add( BitmapFrame.Create( frame, null, metadata, null ) );
					encoder.Save( streamOut );
				}
			}
		}

		/// <summary>
		/// GeoTag を新たに設定します。
		/// </summary>
		/// <param name="latitude">緯度。</param>
		/// <param name="longitude">経度。</param>
		public void SetGeoTag( double latitude, double longitude )
		{
			if( this.GeoTag == null )
			{
				this.GeoTag = new GeoTag() { Latitude = latitude, Longitude = longitude };
				this._isEditedGeoTag = true;
			}
			else
			{
				this.GeoTag.Latitude  = latitude;
				this.GeoTag.Longitude = longitude;
			}
		}

		/// <summary>
		/// 現在の Metadata を表す System.String を返します。
		/// </summary>
		/// <returns>GeoTag の内容を表す System.String。</returns>
		public override string ToString()
		{
			if( this.GeoTag != null )
			{
				return ( this.Address == null ? this.GeoTag.ToString() : String.Format( "{0}、{1}", this.GeoTag.ToString(), this.Address.ToString() ) );
			}
			else if( this.Address != null )
			{
				return this.Address.ToString();
			}

			return String.Empty;
		}

		/// <summary>
		/// オリジナル画像の撮影日時、デジタル化日時、更新日時のメタデータを書き込みます。
		/// これらデータは原則として変更不可なので、存在しない場合だけ設定します。
		/// <para>
		/// 日時情報はオリジナル画像の撮影日時が基準となるので、このデータが存在しない場合は、
		/// 全てのデータを処理時点の日時に統一して書き込みます。
		/// </para>
		/// </summary>
		/// <param name="metadata">画像のメタデータを表すオブジェクト。</param>
		private void WriteDateTime( BitmapMetadata metadata )
		{
			// オリジナル画像の撮影日時
			var dateTimeOriginal = metadata.GetQuery( MetadataViewModel.ExifPathDateTimeOriginal ) as string;
			if( String.IsNullOrEmpty( dateTimeOriginal ) )
			{
				var now = DateTime.Now.ToString( "yyyy:MM:dd hh:mm:ss" );
				metadata.SetQuery( MetadataViewModel.ExifPathDateTimeOriginal,  now );
				metadata.SetQuery( MetadataViewModel.ExifPathDateTimeDigitized, now );
				metadata.SetQuery( MetadataViewModel.ExifPathDateTime,          now );
			}
			else
			{
				// デジタル化日時が欠落している場合は、撮影日時に合わせる
				var dateTimeDigitized = metadata.GetQuery( MetadataViewModel.ExifPathDateTimeDigitized ) as string;
				if( String.IsNullOrEmpty( dateTimeDigitized ) )
				{
					metadata.SetQuery( MetadataViewModel.ExifPathDateTimeDigitized, dateTimeOriginal );
				}

				// 更新日時が欠落している場合は、撮影日時に合わせる
				var dateTime = metadata.GetQuery( MetadataViewModel.ExifPathDateTime ) as string;
				if( String.IsNullOrEmpty( dateTime ) )
				{
					metadata.SetQuery( MetadataViewModel.ExifPathDateTime, dateTimeOriginal );
				}
			}
		}

		#region プロパティ

		/// <summary>
		/// GeoTag を所有している事を示す値を取得します。
		/// </summary>
		public bool HasGeoTag { get { return ( this._geoTag.Value != null ); } }

		/// <summary>
		/// メタデータが編集された事を示す値を取得します。
		/// </summary>
		public bool IsEdited { get { return ( this._isEditedGeoTag || this._isEditedAddress ); } }

		/// <summary>
		/// Address を取得します。
		/// </summary>
		public Address Address { get { return this._address.Value; } private set { this._address.Value = value; } }

		/// <summary>
		/// GeoTag を取得します。
		/// </summary>
		public GeoTag GeoTag { get { return this._geoTag.Value; } private set { this._geoTag.Value = value; } }

		#endregion

		#region フィールド

		/// <summary>
		/// 住所。
		/// </summary>
		private EditValue< Address > _address = new EditValue< Address >();

		/// <summary>
		/// GeoTag。
		/// </summary>
		private EditValue< GeoTag > _geoTag = new EditValue< GeoTag >();

		/// <summary>
		/// Address の編集が行われた事を示す値。
		/// </summary>
		private bool _isEditedAddress;

		/// <summary>
		/// GeoTag の編集が行われた事を示す値。
		/// </summary>
		private bool _isEditedGeoTag;
		
		#endregion

		#region 定数
		
		/// <summary>
		/// EXIF のバージョン。
		/// 型指定が無いので、バイト配列として書き込みを行う必要がある。
		/// </summary>
		private const string ExifPathVersion = "/app1/ifd/exif/{uint=36864}";

		/// <summary>
		/// EXIF における「画像の変更日時」を示すパス文字列。
		/// データは "YYYY:MM:DD HH:MM:SS" + 0x00 の合計 20 バイトとなる。
		/// 通常は DateTimeOriginal と同じ値が設定される。
		/// </summary>
		private const string ExifPathDateTime = "/app1/ifd/exif:{ushort=306}";

		/// <summary>
		/// EXIF における「オリジナル画像の撮影日時」を示すパス文字列。
		/// データは "YYYY:MM:DD HH:MM:SS" + 0x00 の合計 20 バイトとなる。
		/// </summary>
		private const string ExifPathDateTimeOriginal = "/app1/ifd/exif/{ushort=36867}";

		/// <summary>
		/// EXIF における「オリジナル画像がデジタル化された日時」を示すパス文字列。
		/// データは "YYYY:MM:DD HH:MM:SS" + 0x00 の合計 20 バイトとなる。
		/// </summary>
		private const string ExifPathDateTimeDigitized = "/app1/ifd/exif/{ushort=36868}";

		#endregion
	}
}
